% Copyright (c) 2011 Joris Degroote, Majid Hojjat, Electra Stavropoulou,
% Roland Wüchner, Kai-Uwe Bletzinger
%
% Licensed under the Non-Profit Open Software License version 3.0
% (NPOSL-3.0). You should have received a copy of the NPOSL-3.0 along with 
% this file. If not, see <http://www.opensource.org/licenses/NPOSL-3.0>
%
% Primal and adjoint fluid-structure interaction simulation using 
% the IQN-ILS coupling algorithm.
%
% Reference: J. Degroote, K.-J. Bathe, and J. Vierendeels. Performance of a
% new partitioned procedure versus a monolithic procedure in fluid-
% structure interaction. Computers & Structures, 87(11�12):793�801, 2009. 
%
% DOI: 10.1016/j.compstruc.2008.11.013

function [j,djds,hist]=Simulation(m_e,s,reference,graphics,output,dt, ...
    rho_f,algorithm,reuse)

if (nargin<1)
    m_e=100;
end
if (nargin<2)
    s=zeros(m_e+1,1);
end
if (nargin<3)
    reference=false;
end
if (nargin<4)
    graphics=false;
end
if (nargin<5)
    output=0;
end
if (nargin<6)
    dt=0.01;
end
if (nargin<7)
    rho_f=1060.0;
end
if (nargin<8)
    algorithm='QuasiNewton';
end
if (nargin<9)
    reuse=3;
end

k_min=2;
k_e=25;
n_e=100;
ell=0.126;
dz=ell/m_e;
epsilon=1e-5;
small=1e-16;

fluid=Fluid(dt,dz,m_e,s,rho_f);
structure=Structure(dt,dz,m_e,s);
animator=Animator(dt,dz,m_e,n_e,0.1*dt,graphics);
convergence=Convergence(epsilon,k_min,k_e,n_e,small,output);
costfunction=Costfunction(m_e,s,output,reference);
extrapolator=Extrapolator(fluid.x_sg);
switch lower(algorithm)
    case 'quasinewton'
        model=JacobianInverse(small,reuse);
        omega=0.01;
    case 'gaussseidel'
        model=Dummy();
        omega=1.0;
    otherwise
        error('Unknown algorithm');
end
objects={fluid,structure,animator,convergence,costfunction,...
    extrapolator,model};

for n=1:n_e
    cellfun(@increase_time,objects);
    
    for k=1:k_e
        if (k==1)
            x_sg=extrapolator.predict();
        elseif ((k==2) && (~model.ready()))
            x_sg=x_sg+omega*r;
        else
            dx_sg=model.predict(-r)+r;
            x_sg=x_sg+dx_sg;
        end
        x_fg=fluid.calculate_primal(x_sg);
        xt_sg=structure.calculate_primal(x_fg);
        r=xt_sg-x_sg;
        model.add(r,xt_sg);
        
        convergence.add(r);
        if (convergence.is_satisfied())
            break;
        end
    end
    animator.add(fluid.x_f,structure.x_s);
    costfunction.add(fluid.x_f,structure.x_s,fluid.calculate_gradient(),...
        structure.calculate_gradient());
    extrapolator.add(x_sg);
end;

if ((nargout>0) || (reference))
    cellfun(@increase_time,objects);
    costfunction.finish_primal();
    j=costfunction.calculate_value();
    convergence.finish_primal();
    if (nargout>1)
        extrapolator.finish_primal(fluid.C_stxa_sg);
        model.finish_primal();
        if (nargout>2)
            hist=[animator.time animator.r_history];
        end
    
        for n=n_e:-1:1
            cellfun(@decrease_time,objects);
            fluid.adjust_rhs_dual(...
                costfunction.calculate_gradient_f(),...
                structure.calculate_rhs_dual());
            structure.adjust_rhs_dual(...
                costfunction.calculate_gradient_s(),...
                fluid.calculate_rhs_dual());
        
            for k=1:k_e
                if (k==1)
                    C_stxa_sg=extrapolator.predict();
                elseif ((k==2) && (~model.ready()))
                    C_stxa_sg=C_stxa_sg+omega*r;
                else
                    dC_stxa_sg=model.predict(-r)+r;
                    C_stxa_sg=C_stxa_sg+dC_stxa_sg;
                end
                C_ftxa_fg=fluid.calculate_dual(C_stxa_sg);
                C_stxat_sg=structure.calculate_dual(C_ftxa_fg);
                r=C_stxat_sg-C_stxa_sg;
                model.add(r,C_stxat_sg);
        
                convergence.add(r);
                if (convergence.is_satisfied())
                    break;
                end
            end
            costfunction.add(fluid.a_f,structure.a_s,[],[]);
            extrapolator.add(C_stxa_sg);
        end
    
        costfunction.finish_dual();
        djds=costfunction.calculate_gradient();
        convergence.finish_dual();
    end
end