% Copyright (c) 2011 Joris Degroote, Majid Hojjat, Electra Stavropoulou,
% Roland Wüchner, Kai-Uwe Bletzinger
%
% Licensed under the Non-Profit Open Software License version 3.0
% (NPOSL-3.0). You should have received a copy of the NPOSL-3.0 along with 
% this file. If not, see <http://www.opensource.org/licenses/NPOSL-3.0>
%
% Calculation of the structural motion in a tube.

classdef Structure < handle
    properties (Constant)
        E_o=4.0e5;
        G=4.0e5;
        nu=0.5;
        rho=1000.0;
        r_o=0.003;
        h=0.0003;
    end
    properties (SetAccess=immutable)
        dt;
        dz;
        m_e;
        m_so;
        m_sg;
        m_sh;
        m_fg;
        E;
        pdEpds;
        k;
        s;
        
        C_s;
        D_s;
        M_s;
        N_s;
    end
    properties (SetAccess=private)
        n;
        
        b_s;
        x_s;
        x_sn;
        x_fg;
        x_fgn;
        
        pdjpdx_s;
        a_s;
        a_sn;
        C_ftxa_fg;
        D_ftxa_fgn;
    end
    methods
        function S=Structure(dt,dz,m_e,s)
            S.dt=dt;
            S.dz=dz;
            S.m_e=m_e;
            S.m_so=S.m_e;
            S.m_sg=S.m_e;
            S.m_sh=S.m_e;
            S.m_fg=S.m_e;
            S.E=S.E_o*(1.0+s(1:m_e,1)*0.5);
            S.pdEpds=S.E_o*0.5;
            S.k=2.0*(1.0+S.nu)/(4.0+3.0*S.nu);
            S.s=s;
                        
            S.C_s=-eye(S.m_sh,S.m_fg);
            S.D_s=zeros(S.m_sh,S.m_fg);
            S.M_s=zeros(S.m_sg+S.m_so);
            S.N_s=zeros(S.m_sg+S.m_so);
            S.M_s(1:S.m_sh,1:S.m_sg)=...
                diag(S.E)*S.h/(1-S.nu^2)/S.r_o^2 ...
                +eye(S.m_sg)*2.0*S.k*S.G*S.h/S.dz^2 ...
                -diag(ones(S.m_sg-1,1)*S.k*S.G*S.h/S.dz^2,1) ...
                -diag(ones(S.m_sg-1,1)*S.k*S.G*S.h/S.dz^2,-1);
            S.M_s(1:S.m_sh,S.m_sg+1:S.m_sg+S.m_so)=...
                eye(S.m_sg)*S.rho*S.h/S.dt;
            S.M_s(S.m_sh+1:S.m_sg+S.m_so,1:S.m_sg)=-eye(S.m_sg)/S.dt;
            S.M_s(S.m_sh+1:S.m_sg+S.m_so,S.m_sg+1:S.m_sg+S.m_so)=...
                eye(S.m_so);
            S.M_s(1,1)=S.M_s(1,1)-S.k*S.G*S.h/S.dz^2;
            S.M_s(S.m_e,S.m_e)=S.M_s(S.m_e,S.m_e)-S.k*S.G*S.h/S.dz^2;
            S.N_s(1:S.m_sh,S.m_sg+1:S.m_sg+S.m_so)=...
                eye(S.m_sg)*S.rho*S.h/S.dt;
            S.N_s(S.m_sh+1:S.m_sg+S.m_so,1:S.m_sg)=-eye(S.m_sg)/S.dt;
            
            S.n=0;
            
            S.x_s=zeros(S.m_sg+S.m_so,1);
            S.x_sn=S.x_s;
            S.x_fg=zeros(S.m_fg,1);
            S.x_fgn=S.x_fg;
            
            S.a_s=zeros(S.m_sg+S.m_so,1);
            S.a_sn=S.a_s;
            S.C_ftxa_fg=zeros(S.m_sg,1);
            S.D_ftxa_fgn=S.C_ftxa_fg;
        end
        function x_sg=calculate_primal(S,x_fg)
            S.x_fg=x_fg;
            S.x_s=S.M_s\(S.b_s+S.N_s*S.x_sn...
                +[S.D_s*S.x_fgn; zeros(S.m_sg+S.m_so-S.m_sh,1)]...
                -[S.C_s*S.x_fg; zeros(S.m_sg+S.m_so-S.m_sh,1)]);
            x_sg=S.x_s(1:S.m_sg,1);
        end
        function C_stxa_sg=calculate_dual(S,C_ftxa_fg)
            S.C_ftxa_fg=C_ftxa_fg;
            S.a_s=S.M_s'\(-S.pdjpdx_s'+S.N_s'*S.a_sn...
                +[S.D_ftxa_fgn; zeros(S.m_sg+S.m_so-S.m_sg,1)]...
                -[S.C_ftxa_fg; zeros(S.m_sg+S.m_so-S.m_sg,1)]);
            C_stxa_sg=S.C_s'*S.a_s(1:S.m_sh,1);
        end
        function D_stxa_sgn=calculate_rhs_dual(S)
            D_stxa_sgn=S.D_s'*S.a_sn(1:S.m_sh);
        end
        function adjust_rhs_dual(S,pdjpdx_s,D_ftxa_fgn)
            S.pdjpdx_s=pdjpdx_s;
            S.D_ftxa_fgn=D_ftxa_fgn;
        end
        function pdR_spds=calculate_gradient(S)
            pdR_spds=zeros(S.m_sg+S.m_so,length(S.s));
            pdR_spds(1:S.m_sh,1:length(S.s)-1)=...
                S.pdEpds*S.h/(1.0-S.nu^2)/S.r_o^2*diag(S.x_s(1:S.m_sg));
        end
        function increase_time(S)
            S.n=S.n+1;
            
            S.b_s=zeros(S.m_sg+S.m_so,1);
            S.x_sn=S.x_s;
            S.x_fgn=S.x_fg;
        end
        function decrease_time(S)
            S.n=S.n-1;
            
            S.pdjpdx_s=[];
            S.a_sn=S.a_s;
            S.D_ftxa_fgn=[];
        end
    end
end